// Sleeper Cell Alarm
// Copyright (c) 2016, 2017, 2018, 2019, 2020 by Haute Solutions / Ed Van Every (www.haute-solutions.com)
//
// FUNCTION: This sketch will implement the Sleeper Cell Alarm functionality on the associated hardware.
//           This is a full featured Vehicle Alarm System providing tradional Alarm Alerts (Lights, Siren) as well
//           as SMS notification with GPS tracking.  The Sleeper Cell Alarm also draws absolutely no power until
//           tripped. (Please refer to the Sleeper Cell Alarm Documentation for further information)
//
// HARDWARE: This sketch requires the following Hardware:
//
//           Haute Solutions MeGGaBoard (2G/3G) with Haute Solutions SleeperCell Alarm Shield (V1.0)
//           -or-
//           Haute Solutions ESP324G with Haute Solutions SleeperCell Alarm Shield (V2.0)
//
// SOFTWARE: This sketch requires the following Libraries and Software:
//
//           Haute Solutions GSMSMS Library
//
// LICENSE:  This program is free software: you can redistribute it and/or modify
//           it under the terms of the GNU General Public License as published by
//           the Free Software Foundation, either version 3 of the License, or
//           (at your option) any later version.
//
//           This program is distributed in the hope that it will be useful,
//           but WITHOUT ANY WARRANTY; without even the implied warranty of
//           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//           GNU General Public License for more details.
//           <http://www.gnu.org/licenses/>.
//
// ----------------------------- Sleeper Cell Alarm -----------------------------------------------------------------------------------------------

#define VERSION "V3.1"                   // Firmware Version Info

#define GOOGLE_MAPS_LOCATOR "https://www.google.com/maps/search/?api=1&query="

// Persistant Storage Info (EEPROM Info): Character Data, NULL Terminated
// All Fields are 20 Chars max (Inc NULL Terminator)
// The values below are the Field Offsets into EEPROM memory for start of each string
// Currently (V1 EEPROM Layout) Uses 220 Bytes Total

#define EESIG           0                // EEPROM Data Valid Signature ("SMSRELAY EEPROM V1")
#define EEALARMID       1                // Alarm ID ("Trailer") - For Multiple Alarms
#define EESILENT        2                // Silent Alarm:  "0" or "1"
#define EEDELAY         3                // Alarm Delay: "0" or "1"
#define EERESET         4                // Alarm Reset Time: {# of Minutes, 0 for No Reset}
#define EEGPSUPDATE     5                // GPS Update Interval: {# of Minutes, 0 for GPS Disable}
#define EENOTIFY        6                // Start of SMS Notification List

#define EEPROMSig "SLEEPERCELL V1.0"     // EEPROM Valid Signature (Version is EEPROM specific - NOT Firmware Specific)
#define EEFIELDSIZE     20               // Size of each field/string (inc Terminator)
#define EEMAXSMS        5                // Max Number of SMS Notification Numbers

#define SMSDELAY        1                // Number of seconds to wait between SMS Messages

#define BUFFSIZE 512
#define NUMSIZE 20

#define NUMCFGCMDS 16
const char *ConfigCmdArray[NUMCFGCMDS] = { 
  "ALARMID:",
  "SILENT:",
  "DELAY:",
  "RESET:",
  "GPSUPDATE:",
  "NOTIFY:",
  "CLEARNOTIFY",
  "SHOWNOTIFY",
  "STATUS",
  "HELP",
  "CLEAREEPROM",     //Undocumented
  "SIMINFO",         //Undocumented
  "MEINFO",          //Undocumented
  "CELLINFO",        //Undocumented
  "OPERATORS",       //Undocumented
  "ATCMD:"           //Undocumented
};

#define NUMALARMCMDS 4
const char *AlarmCmdArray[NUMALARMCMDS] = {
  "QUIET",
  "RESET",
  "HELP",
  "GPSLOC"
};

// --------------------------------------------------------------------------------------------------------------------------------------------------

#include <EEPROM.h>
#include <GSMSMS.h>                      // GSM SMS Library

// Prep for ESP324G Board
#if defined(ESP32)
#define CONFIG_PIN 16
#define GSM_PIN 17
#define STATUS_PIN 32
#define RESET_PIN 25
#define SIREN_PIN 26
#define LIGHTS_PIN 27
GSMSMS gsmsms(ESP324G);
#endif

// Prep for MEGGaBoard
#if defined(__AVR_ATmega2560__)
#define CONFIG_PIN 4
#define GSM_PIN 5
#define STATUS_PIN 6
#define RESET_PIN 7
#define SIREN_PIN 8
#define LIGHTS_PIN 9
GSMSMS gsmsms(MEGGABOARD3G);
#endif

// Globals ------------------------------------------------------------------------------------------------------------------------------------------

// Globals from EEPROM
int Silent;
int Delay;
int Reset;
int GPSUpdate;
char AlarmID[20];

char Buffer[BUFFSIZE];                   // Primary work Buffer
char Caller[NUMSIZE];
int Config;
int Siren;
int Lights;
int Isolated;
int SMSSent;
int ShowStarted;
int CanNotify;
int SIM;
char Latitude[20];
char Longitude[20];
char Temp[200];
int Dots;
int CSQ;

unsigned long ResetTimer;
unsigned long GPSTimer;

typedef struct {
  int led;
  unsigned long msdelay;
  unsigned long timeref;
  int blinks;
  int state;
} BLINKSTRUCT;

BLINKSTRUCT StatusBlinker;              // Status LED Control Structure
BLINKSTRUCT LightsBlinker;              // Lights Relay Control Structure

int RTBlinker(BLINKSTRUCT *blinkstruct, int blinkcount = -3, unsigned long msdelay = 0);

// Code...

void setup() {
int i;

#if defined(ESP32)
  EEPROM.begin((EENOTIFY + EEMAXSMS) * EEFIELDSIZE);
#endif

  Serial.begin(9600);
  Serial.println();
  Serial.println();

  Serial.print("Sleeper Cell ");
  Serial.print(VERSION);
  
  Serial.println();
  Serial.println();

  SetupPins();

  RTBlinker(&StatusBlinker, STATUS_PIN, 500);
  RTBlinker(&LightsBlinker, LIGHTS_PIN, 1000);

  Serial.print("EEPROM ");
  if(IsEEPROMValid()) Serial.println("Valid");
  else {
    InitializeEEPROM();
    Serial.println("Initialized");
  }

  if(digitalRead(CONFIG_PIN) == HIGH) Config=1;
  else Config=0;

  if(digitalRead(GSM_PIN) == HIGH) Isolated=0;  // GSM ON = !Isolated
  else Isolated=1; 

  // Manual Override for testing
  // Config=0;
  // Isolated=0;
  //

  if(Config) Isolated=0;                        // Override Isolation if Config = ON

  if(!Isolated) ReadEEPROMData();
  else {
    Silent=0;
    Delay=0;
    Reset=5;
    GPSUpdate=0;
  }

  CSQ = 0;
  SMSSent = 0;
  ShowStarted = 0;
  Dots = 0;

  //  Start the SHOW ASAP if appropriate
  if(Delay+Silent+Config == 0) StartTheShow();

  DisplayConfig();

  if(!Isolated) {
    
    // Initialize GSM Chip and set Operating Parameters
    Serial.println("GSM Init...");
    gsmsms.Init();
    Serial.println("GSM SetParams...");
    gsmsms.SetParams();
    gsmsms.SendCmd("AT+CNMI=2,1");              // Enable Unsolicited SMS Notification
    gsmsms.SendCmd("AT+AUTOCSQ=1,1");           // Enable Unsolicited CSQ Notification (whenever it changes)
//  gsmsms.SendCmd("AT+AUTOCSQ=1,0");           // Enable Unsolicited CSQ Notification (every 5 secs)
    gsmsms.DeleteAllSMS();
    Serial.println("GSM Ready!");

    DisplayHardwareInfo();

    if(GPSUpdate) {
      gsmsms.EnableGPS();
      Serial.println("GPS Started");
    }
    
    CanNotify=0;
    if(!gsmsms.SIMInserted()) SIM=0;            // No SIM Detected!
    else {
      SIM=1;
      CSQ=gsmsms.GetSignalQuality();
      for(i=0; i<EEMAXSMS; i++) {
        if(strlen(ReadEEPROMSMSNumber(i+1,Temp))) {
          CanNotify=1;
          if(Config) SendConfig(Temp, "Config Mode Enabled:");
          delay(SMSDELAY*1000);
        }
      }
    }

    CheckStatus();
  }

  delay(5000);
  
  Serial.println("Ready!");
  Serial.println();
  Serial.flush();
}

void SetupPins() {

  // Config Mode Sense Pin
  pinMode(CONFIG_PIN, INPUT);
  pinMode(GSM_PIN, INPUT);
  
  // Relay Pins
  pinMode(RESET_PIN, OUTPUT);
  pinMode(SIREN_PIN, OUTPUT);
  pinMode(LIGHTS_PIN, OUTPUT);

  // Indicator LED Pin
  pinMode(STATUS_PIN, OUTPUT);

  // Relay Pins must be set HIGH to disable relay
  digitalWrite(RESET_PIN,HIGH);
  digitalWrite(SIREN_PIN,HIGH);
  digitalWrite(LIGHTS_PIN,HIGH);  
}

void loop() {
static unsigned long MinuteTimer = 0;
unsigned long TimeElapsed;

  if(!Config) AlarmMode();

  RTBlinker(&StatusBlinker);                  // Service the Status LED 
  RTBlinker(&LightsBlinker);                  // Service the Lights Relay 

  ProcessUnsolicitedMsg();                    // Any Messages?

  ProcessConsoleCommand();                    // Any Console Commands?

  TimeElapsed = millis() - MinuteTimer;
  if( TimeElapsed > 60000l ) {                // Failsafe Polling Operations every minute

    CheckSMS();
    CheckStatus();
    MinuteTimer = millis();

  }
}

void AlarmMode() {

  if(!Isolated) {
    if(!SMSSent) {
      if(GPSUpdate) {    // Get out an alarm notification ASAP (without waiting for GPS Fix...)
        sprintf(Buffer,"%s Alarm Activated!\n\n(Acquiring GPS Fix...)", AlarmID);
        SendAlarmSMS(Buffer);
        SMSSent++;
      }
      SendAlarmNotification(GPSUpdate);
    }

    if(GPSUpdate) {
      if( millis() > (GPSTimer + (long(GPSUpdate)*60000l)) ) {
        SendAlarmNotification(GPSUpdate);
        GPSTimer=millis();
      }
    }
  }

  if(!Silent) {
    if(!ShowStarted) StartTheShow();
  }
  
  if(Reset) {
    if(millis() > (ResetTimer + (long(Reset)*60000l)) ) ShutDown();
  }
}

void ProcessUnsolicitedMsg() {
int c;
int i;

  if(!gsmsms.Available()) return;

  Buffer[0]='\0';
  while(gsmsms.Available()) {
    c = gsmsms.Read(); 
    if ((i = strlen(Buffer)) < (BUFFSIZE - 1)) {
      Buffer[i] = c;  
      Buffer[i+1] = '\0';
    }
    if(!gsmsms.Available()) delay(100);
  }

  Serial.print("\n\n");
  Serial.println(Buffer);   // Local Echo

  if(strstr(Buffer,"+CMTI:")) {
    CheckSMS();
    return;
  }

  if(strstr(Buffer,"+CSQ:")) {
    CheckGSM(Buffer);
    return;
  }

  return;
}

void CheckSMS() {

  if(gsmsms.GetSMS(Caller, Buffer, NUMSIZE, BUFFSIZE) == 0) return;

  Serial.println();
  Serial.print("(");
  Serial.print(Caller);
  Serial.print(") ");
  Serial.println(Buffer);

  if(Config) ProcessConfigCommand();
  else ProcessAlarmCommand();

  return;
}

void CheckGSM(char *buff) {
char *p1;
char *p2;
int csq;

  p1=strstr(buff,"+CSQ:");
  if(p1 == NULL) return;
  p1+=6;

  p2=strchr(p1,',');
  if(p2 == NULL) return;
  *p2='\0';

  csq=atoi(p1);
  
  CSQ = 0;
  if(csq > 1)  CSQ = 1;
  if(csq > 5)  CSQ = 2;
  if(csq > 9)  CSQ = 3;
  if(csq > 14) CSQ = 4;
  if(csq > 19) CSQ = 5;
  if(csq == 99) CSQ=0;
}

void ProcessConsoleCommand() {
int i;
char c;

  if(!Serial.available()) return;

  Buffer[0]='\0';
  while(Serial.available()) {
    c = Serial.read(); 
    if ((i = strlen(Buffer)) < (BUFFSIZE - 1)) {
      Buffer[i] = c;  
      Buffer[i+1] = '\0';
    }
    if(!Serial.available()) delay(100);
  }

  Serial.print("\n\n");
  Serial.print(Buffer);   // Local Echo

  gsmsms.SendCmdParams(Buffer,30000,1000,"OK",10);
  strncpy(Buffer,gsmsms.GSMCommBuff,BUFFSIZE);
  Buffer[BUFFSIZE-1]='\0';

  Serial.println(Buffer);
  Dots=0;
}

void CheckStatus() {
int FlashCount;

//List of Flash Codes
//
//  0 = No Service (LED Off)
// -2 = Service OK (LED On)
//  2 = No Recipient(s) Configured
//  3 = SIM Not Detected
//  4 = Network Registration Failed 

  if(Isolated) FlashCount=0;                          // No Service
      
  else if(!SIM) FlashCount=3;                         // No SIM Inserted
  
  else {
    switch (gsmsms.RegisteredOnNetwork()) {
      case 0:                                         // Not Registered
        FlashCount=4;
        break;
      case 1:                                         // Registered on Home Network (OK)
        FlashCount=-2;
        break;
      case 2:                                         // Not Registered, still searching
        FlashCount=0;
        break;
      case 3:                                         // Registration Denied
        FlashCount=4;
        break;
      case 4:                                         // Unknown
        FlashCount=4;
        break;
      case 5:                                         // Registered, Roaming (OK)
        FlashCount=-2;
        break;
    }
  }

  if ((FlashCount == -2) && (CanNotify == 0)) FlashCount=2;  // No Recipient(s) configured

  if(StatusBlinker.blinks == FlashCount) return;             // No Change;
  
  RTBlinker(&StatusBlinker, FlashCount);                     // Update the Status  
}

void DisplayConfig() {
int i;

  Serial.println("");

  Serial.print("AlarmID: ");
  Serial.println(AlarmID);

  Serial.print("Silent: ");
  Serial.println(Silent);

  Serial.print("Delay: ");
  Serial.println(Delay);

  Serial.print("Reset: ");
  Serial.println(Reset);

  Serial.print("GPSUpdate: ");
  Serial.println(GPSUpdate);

  for(i=0; i<EEMAXSMS; i++) {
    sprintf(Buffer,"SMSNotify %d: %s",i+1,ReadEEPROMSMSNumber(i+1,Temp));
    Serial.println(Buffer);
  }

  Serial.print("Config: ");
  Serial.println(Config);

  Serial.print("Isolated: ");
  Serial.println(Isolated);

  Serial.println("");
}

void DisplayHardwareInfo() {
int i;

  Serial.println("");

  gsmsms.GetChipset(Temp);
  Serial.print("Chipset: ");
  Serial.println(Temp);

  gsmsms.GetIMEI(Temp);
  Serial.print("IMEI: ");
  Serial.println(Temp);

  if(gsmsms.SIMInserted()) gsmsms.GetICCID(Temp);
  else strcpy(Temp,"No SIM");
  Serial.print("ICCID: ");
  Serial.println(Temp);

  if(gsmsms.SIMInserted()) gsmsms.GetIMSI(Temp);
  else strcpy(Temp,"No SIM");
  Serial.print("IMSI: ");
  Serial.println(Temp);

  if(gsmsms.SIMInserted()) gsmsms.GetSubscriberNumber(Temp);
  else strcpy(Temp,"No SIM");
  Serial.print("Number: ");
  Serial.println(Temp);

  Serial.println("");
}

void SendAlarmNotification(int gps) {

  sprintf(Buffer,"%s Alarm Activated!", AlarmID);

  if(gps) {
    GPSTimer=millis();
    AppendGPSInfo(Buffer);
  }

  SendAlarmSMS(Buffer);

  SMSSent++;
}

int AppendGPSInfo(char *p) {
char *p1;

  if(!gsmsms.GetGPSPosition(Latitude, Longitude)) strcat(p,"Lat:  N/A\nLong: N/A");
  else {
    strcat(p,"\n\n");
    strcat(p,"Lat:  ");
    strcat(p,Latitude);
    strcat(p,"\nLong: ");
    strcat(p,Longitude);
    strcat(p,"\n\n");
    strcat(p,GOOGLE_MAPS_LOCATOR);
    strcat(p,Latitude);
    strcat(p,",");
    strcat(p,Longitude);
  }
  return 1;
}

void SendAlarmSMS(const char *msg) {
int i;

  for(i=0; i<EEMAXSMS; i++) {
    ReadEEPROMSMSNumber(i+1,Temp);
    if(strlen(Temp)) {
      gsmsms.SendSMS(Temp,msg);
      delay(SMSDELAY);
    }
  }
}

void StartTheShow() {

  Serial.println("StartTheShow!");

  ShowStarted=1;
  Silent=0;
  
  SirenOn();
  RTBlinker(&LightsBlinker, -1);

  ResetTimer=millis();
}

void StopTheShow()
{
  Serial.println("StopTheShow!");

  ShowStarted=0;
  Silent=1;

  SirenOff();
  RTBlinker(&LightsBlinker, 0);

  ResetTimer=0l;
}

void ProcessConfigCommand() {
char *p1;
char *param;
int CmdID;

  if(strlen(ReadEEPROMSMSNumber(1,Temp)) == 0) {
    if(strlen(Caller) > 4) {
      WriteEEPROMSMSNumber(1,Caller);
      CanNotify=1;
      SendNotificationList(Caller);
    }
  }

  CmdID = GetCommandID(Buffer, ConfigCmdArray, NUMCFGCMDS);
  param = strchr(Buffer, ':');
  if(param != NULL) param = param + strspn(param, " :");
  
  switch (CmdID) {

    case 0:     // Not Valid
      if(!CallerAuthorized(Caller)) {
        Serial.println();
        Serial.print(Caller);
        Serial.print(" Not Authorized! (");
        Serial.print(Buffer);
        Serial.println(")");
        ForwardMsgToAll();
      } else {
        gsmsms.SendSMS(Caller,"What?");
      }
      return;  

    case 1:     // AlarmID:
      strncpy(AlarmID,param,20);
      AlarmID[19]='\0';
      WriteEEPROMField(EEALARMID,AlarmID);
      SendConfig(Caller, NULL);;
      return;

    case 2:     // Silent:
      if(!strcasecmp(param,"YES")) {
        Silent=1;
        Reset=0;
      } else if(!strcasecmp(param,"NO")) {
        Silent=0;
        Reset=5;
      } else {
        gsmsms.SendSMS(Caller,"What?");
        return;  
      }
      WriteEEPROMIntField(EESILENT, Silent);
      WriteEEPROMIntField(EERESET, Reset);
      SendConfig(Caller, NULL);;
      return;

    case 3:     // Delay:
      if(!strcasecmp(param,"YES")) Delay=1;
      else if(!strcasecmp(param,"NO")) Delay=0;
      else {
        gsmsms.SendSMS(Caller,"What?");
        return;  
      }
      WriteEEPROMIntField(EEDELAY,Delay);
      SendConfig(Caller, NULL);;
      return;

    case 4:     // Reset:
      Reset=atoi(param);
      WriteEEPROMIntField(EERESET,Reset);
      SendConfig(Caller, NULL);;
      return;

    case 5:     // GPSUPDATE:
      GPSUpdate=atoi(param);
      WriteEEPROMIntField(EEGPSUPDATE,GPSUpdate);
      SendConfig(Caller, NULL);;
      return;

    case 6:     // NOTIFY:
      if(!strcasecmp(param,"ME")) strcpy(Temp,Caller);
      else strcpy(Temp,param);    
      if(!SMSNumberValid(Temp)) gsmsms.SendSMS(Caller,"Invalid Format! Use \"+12625551212\" or \"2625551212\"");   
      else if(!AddSMSNotification(Temp)) gsmsms.SendSMS(Caller,"Notification List Full! (\"CLEARNOTIFY\" to clear)");
      else SendNotificationList(Caller);
      return;

    case 7:     // CLEARNOTIFY
      ClearSMSNotificationList();
      SendNotificationList(Caller);
      CanNotify=0;
      CheckStatus();
      return;

    case 8:     // SHOWNOTIFY
      SendNotificationList(Caller);
      return;

    case 9:     // STATUS
      SendConfig(Caller, NULL);
      return;

    case 10:    // HELP
      SendConfigHelp(Caller);
      return;

    case 11:    // CLEAREEPROM
      InitializeEEPROM();
      CanNotify=0;
      CheckStatus();
      SendConfig(Caller, "EEPROM Initialized");
      return;

    case 12:    // SIMInfo (Undocumented)
      SendSIMInfo(Caller);
     return;

    case 13:    // MEInfo (Undocumented)
      SendMEInfo(Caller);
      return;

    case 14:    // CellInfo (Undocumented)
      SendCellInfo(Caller,gsmsms.GetSignalQuality());
      return;

    case 15:    // Operators (Undocumented)
      SendOperators(Caller);
      return;

    case 16:    // ATCMD (Undocumented)
      p1=strchr(param,'\n');
      if(p1 != NULL) *p1='\0';
      gsmsms.SendCmdParams(param,30000,1000,"OK",10);
      strncpy(Buffer,gsmsms.GSMCommBuff,BUFFSIZE);
      Buffer[BUFFSIZE-1]='\0';
      gsmsms.SendSMS(Caller,Buffer);
      return;

  }  
}

void ProcessAlarmCommand() {
char *param;
int CmdID;

  if(!CallerAuthorized(Caller)) {
    Serial.println();
    Serial.print(Caller);
    Serial.print(" Not Authorized! (");
    Serial.print(Buffer);
    Serial.println(")");
    return;
  }


  CmdID = GetCommandID(Buffer, AlarmCmdArray, NUMALARMCMDS);
  param = strchr(Buffer, ':');
  if(param != NULL) param = param + strspn(param, " :");

  switch (CmdID) {

    case 0:     // {ECHO to all recipients}
      ForwardMsgToAll();
      return;

    case 1:     // QUIET
      StopTheShow();
      gsmsms.SendSMS(Caller,"OK! (Quiet)");
      return;

    case 2:     // RESET
      ShutDown();                                       // Should never return from here unless a trigger remains closed
      gsmsms.SendSMS(Caller,"Unable! (Trigger Closed)");
      return;

    case 3:     // HELP
      SendAlarmHelp(Caller);
      return;

    case 4:     // GPSLOC
      SendAlarmNotification(1);
      return;

  }
}

void ForwardMsgToAll() {
int i;
int pushval;
int msglen;
int oldloc;
int newloc;

  pushval = (6 + strlen(Caller) + 2);
  Buffer[BUFFSIZE-pushval-1]='\0';
  msglen = strlen(Buffer);
  for(i=msglen; i>0; i--) {
    oldloc=i-1;
    newloc=oldloc+pushval;
    Buffer[newloc] = Buffer[oldloc];
  }
  strncpy(Buffer,"(From ",6);
  strncpy(Buffer+6,Caller,strlen(Caller));
  strncpy(Buffer+6+strlen(Caller),") ",2); 
  
  SendAlarmSMS(Buffer);
}

int GetCommandID(const char *astring, const char *CmdArray[], int NumCmds) {
int i;
 
  for (i=0; i<NumCmds; i++) {
    if(!strncasecmp(astring,CmdArray[i],strlen(CmdArray[i]))) return i+1;
  }
  return 0;
}

void SendSIMInfo(const char *acaller){
char *p1;
char *p2;
char *p3;

  strcpy(Buffer,"");

  // ICCID
  gsmsms.GetICCID(Temp);
  strcat(Buffer,"ICCID: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
  
  // IMSI
  gsmsms.GetIMSI(Temp);
  strcat(Buffer,"IMSI: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  // Service Provider
  gsmsms.GetATResult("AT+CSPN?","+CSPN: ",Temp,1000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"Provider: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");
  
  // Subscriber Number
  gsmsms.GetSubscriberNumber(Temp);
  strcat(Buffer,"Number: ");
  strcat(Buffer,Temp);
  
  gsmsms.SendSMS(acaller,Buffer);
}

void SendMEInfo(const char *acaller) {
  
  strcpy(Buffer,"");

  // Chipset ID
  gsmsms.GetChipset(Temp);
  strcat(Buffer,"Chipset: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
  
  // IMEI
  gsmsms.GetIMEI(Temp);
  strcat(Buffer,"IMEI: ");
  strcat(Buffer,Temp);

  gsmsms.SendSMS(acaller,Buffer);  
}

void SendCellInfo(const char *acaller, int csq) {
char *p1;
char *p2;
char *p3;

  strcpy(Buffer,"");

  // Current Band
  switch(gsmsms.GetSimCom()) {
    case 808:
    case 900:
      gsmsms.GetATResult("AT+CBAND?","+CBAND: ",Temp,1000);
      break;
    case 5320:
      gsmsms.GetATResult("AT*CNTI=0","*CNTI: 0,",Temp,1000);
      break;
  }
  strcat(Buffer,"Band: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  // Signal Quality
  sprintf(Temp,"Signal: %c/5",char(csq+48));
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
 
  // Current Network Operator and Mode
  gsmsms.GetATResult("AT+COPS?","+COPS: ",Temp,10000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"Current: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");
  
  // Service Center Address
  gsmsms.GetATResult("AT+CSCA?","+CSCA: ",Temp,1000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"SMSC: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");

  gsmsms.SendSMS(acaller,Buffer);  
}
 
void SendOperators(const char *acaller) {

  strcpy(Buffer,"");

  // Network Operators Currently in the Network
  gsmsms.GetATResult("AT+COPS=?","+COPS: ",Temp,60000);
  strcat(Buffer,"Operators: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  gsmsms.SendSMS(acaller,Buffer);  
}

int SendConfigHelp(const char *acaller) {
const char msgend[2]={0x1a,'\0'};

  Serial.print("Sending Config Help to: ");
  Serial.println(acaller);

  //  We are going to build our SMS Message incrementally using the low level library functions
  //  So we don't have to allocate a bunch of memory for a temp string

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;

  gsmsms.Print("ALARMID: {text}\n");
  gsmsms.Print("SILENT: [YES|NO]\n");
  gsmsms.Print("DELAY: [YES|NO]\n");
  gsmsms.Print("RESET: {Minutes}\n");
  gsmsms.Print("GPSUPDATE: {Minutes}\n");
  gsmsms.Print("NOTIFY: {SMS Number}\n");
  gsmsms.Print("SHOWNOTIFY\n");
  gsmsms.Print("CLEARNOTIFY\n");
  gsmsms.Print("STATUS\n");
 
  gsmsms.Print(msgend);
  if(gsmsms.FlushSMS() > 0) return 2;
  
  return 0;
}

int SendAlarmHelp(const char *acaller) {
const char msgend[2]={0x1a,'\0'};

  Serial.print("Sending Alarm Help to: ");
  Serial.println(acaller);

  //  We are going to build our SMS Message incrementally using the low level library functions
  //  So we don't have to allocate a bunch of memory for a temp string

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;

  gsmsms.Print("QUIET\n");
  gsmsms.Print("RESET\n");
  gsmsms.Print("GPSLOC\n");
  gsmsms.Print("TELLALL: {text}\n");
 
  gsmsms.Print(msgend);
  if(gsmsms.FlushSMS() > 0) return 2;
  
  return 0;  
}

int AddSMSNotification(const char *num) {
int i;
char anum[20];

  for(i=0; i<EEMAXSMS; i++) {
    ReadEEPROMSMSNumber(i+1,anum);
    if(!strcmp(anum,num)) return 2;
    if(!strlen(anum)) break;
  }
  if(i == EEMAXSMS) return 0;

  if(num[0] == '+') WriteEEPROMSMSNumber(i+1, num);
  else {
    sprintf(anum,"+1%s",num);
    WriteEEPROMSMSNumber(i+1, anum);
  }
  
  return 1;
}

int SMSNumberValid(const char *num) {
  
  if(strlen(num) < 10) return 0;
  if(strchr(num,'-')) return 0;

  if((strlen(num) > 10) && (num[0] != '+')) return 0;

  return 1;
}

void ClearSMSNotificationList() {
int i;

  for(i=0; i<EEMAXSMS; i++) WriteEEPROMSMSNumber(i+1,"");
}

int CallerAuthorized(char *num) {
int i;

  for(i=0; i<EEMAXSMS; i++) {
    ReadEEPROMSMSNumber(i+1,Temp);
    if(!strcmp(Temp, num)) return 1;
  }
  return 0; 
}

int SendConfig(const char *acaller, const char *msg)
{
int  i;
const char msgend[2]={0x1a,'\0'};

  Serial.print("Sending Config Info to: ");
  Serial.println(acaller);

  //  We are going to build our SMS Message incrementally using the low level library functions
  //  So we don't have to allocate a bunch of memory for a temp string

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;
  
  if(msg != NULL) {
    gsmsms.Print(msg);
    gsmsms.Print("\n");
    gsmsms.Print("\n");
  }

  gsmsms.Print("Firmware: ");
  gsmsms.Print(VERSION);
  gsmsms.Print("\n");

  gsmsms.Print("ALARMID: ");
  gsmsms.Print(AlarmID);
  gsmsms.Print("\n");

  gsmsms.Print("SILENT: ");
  if(Silent) gsmsms.Print("Yes");
  else gsmsms.Print("No");
  gsmsms.Print("\n");
  
  gsmsms.Print("DELAY: ");
  if(Delay) gsmsms.Print("Yes");
  else gsmsms.Print("No");
  gsmsms.Print("\n");
  
  gsmsms.Print("RESET: ");
  gsmsms.Print(Reset);
  gsmsms.Print(" Minutes");
  gsmsms.Print("\n");

  gsmsms.Print("GPSUPDATE: ");
  gsmsms.Print(GPSUpdate);
  gsmsms.Print(" Minutes");
  gsmsms.Print("\n");

  gsmsms.Print("Signal: ");
  gsmsms.Print(CSQ);
  gsmsms.Print("/5");
  gsmsms.Print("\n");

  gsmsms.Print(msgend);
  if(gsmsms.FlushSMS() > 0) return 2;

  return 0;
}

int SendNotificationList(const char *acaller)
{
int  i;
const char msgend[2]={0x1a,'\0'};

  //  We are going to build our SMS Message incrementally using the low level library functions
  //  So we don't have to allocate a bunch of memory for a temp string

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;

  for(i=0; i<EEMAXSMS; i++) {
    gsmsms.Print("NOTIFY ");
    gsmsms.Print(i+1);
    gsmsms.Print(": ");
    gsmsms.Print(ReadEEPROMSMSNumber(i+1,Temp));
    gsmsms.Print("\n");
  }
  
  gsmsms.Print(msgend);
  if(gsmsms.FlushSMS() > 0) return 2;
  
  return 0;
}

void SirenOn() {
  digitalWrite(SIREN_PIN,LOW);
  Siren=1;  
}

void SirenOff() {
  digitalWrite(SIREN_PIN,HIGH);
  Siren=0;  
}

void ShutDown()
{

  Serial.println("Shutdown!");

  // Setting relay pin to LOW will open the relay
  
  digitalWrite(RESET_PIN,LOW);
  delay(1000);

  // If the alarm actually shuts down (no triggers keeping it open) we should never reach this point...

  StopTheShow();

  do {} while(1);
}

// ------------------------------------------------------- BLINKER Function -------------------------------------------------------------------------

int RTBlinker(BLINKSTRUCT *blinkstruct, int p2, unsigned long msdelay) {
int blinkcount;

// if msdelay is specified (!= 0) then p2=led pin number for initialization
// else p2=blinkcount for control

  if(msdelay == 0l) blinkcount=p2;            // Control Call
  else {                                      // Initialization Call
    blinkstruct->led = p2;
    if(blinkstruct->led >= 0) pinMode(blinkstruct->led, OUTPUT);
    blinkstruct->msdelay = msdelay;
    blinkstruct->timeref = 0;
    blinkstruct->blinks = 0;
    blinkstruct->state = -1;
    return 0;
  }

// blinkcount Control:
//
//  -3 = Continue Blinking / Processing
//  -2 = Solid On (No Blinking)
//  -1 = Set Continuous Blink (no blink pattern)
//   0 = Stop Blinking
//  1+ = Set Blink Pattern

// Return Codes:
//
// Low -> High = 1  (State Change)
// High -> Low = 2  (State Change)
// High = -1        (No Change)
// Low = -2         (No Change)
// Idle = 0
// Sequence Completed = 3

int elapsed;
  
  switch(blinkcount) {
    case -3:                  // Continue Blinking/Processing
      if(blinkstruct->state < 0) return 0;
      break;
    case -2:                   // Set Solid On
      blinkstruct->blinks = blinkcount;
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, HIGH);
      blinkstruct->timeref = 0;
      blinkstruct->state = 1;
      return 1;
    case 0:                   // Stop Blinking / Turn Off
      blinkstruct->blinks = blinkcount;
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, LOW);
      blinkstruct->timeref = 0;
      blinkstruct->state = -1;
      return 2;
    default:                  // Set Repeating Blink (>0) or Continuous Blink (-1)
      blinkstruct->blinks=blinkcount;
      blinkstruct->timeref = millis();
      blinkstruct->state=0;
      return 1;
  }

  if(blinkstruct->blinks == -2) return -1;

  elapsed = int(((millis()-blinkstruct->timeref))/blinkstruct->msdelay);      // Elapsed Time Units

  if((elapsed < (blinkstruct->blinks*2)) || (blinkstruct->blinks == -1)) {
    if (elapsed % 2) {                            // Odd (Off)
      if(blinkstruct->state == 1) {
        if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, LOW);
        blinkstruct->state=0;
        return 2;
      }
      return -1;
    } else {                                      // Even (On)
      if(blinkstruct->state == 0) {
        if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, HIGH);
        blinkstruct->state=1;
        return 1;
      }  
      return -2;
    }
  } else {
    if(blinkstruct->state == 1) {
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, LOW);
      blinkstruct->state=0;
      return 2;
    }
    if(elapsed >= ((blinkstruct->blinks*2)+2)) {
      blinkstruct->timeref=millis();
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, HIGH);
      blinkstruct->state=1;
      return 3;
    } 
    return -2;
  }
}

// ------------------------------------------------------- EEPROM Functions -------------------------------------------------------------------------

void ReadEEPROMData() {
  ReadEEPROMField(EEALARMID,AlarmID);
  Silent=ReadEEPROMIntField(EESILENT);
  Delay=ReadEEPROMIntField(EEDELAY);
  Reset=ReadEEPROMIntField(EERESET);
  GPSUpdate=ReadEEPROMIntField(EEGPSUPDATE);
}

void WriteEEPROMData() {
  WriteEEPROMField(EEALARMID,AlarmID);
  WriteEEPROMIntField(EESILENT,Silent);
  WriteEEPROMIntField(EEDELAY,Delay);
  WriteEEPROMIntField(EERESET,Reset);
  WriteEEPROMIntField(EEGPSUPDATE,GPSUpdate);
}

int IsEEPROMValid() {

  ReadEEPROMField(EESIG, Buffer);
  Serial.println(Buffer);
  if(!strcmp(Buffer, EEPROMSig)) return 1;
  return 0;
}

void InitializeEEPROM() {
int i;
  
  WriteEEPROMField(EESIG, EEPROMSig);
  
  // Defaults for Connected Alarm (with SMS)
  strcpy(AlarmID, "SleeperCell");
  Silent=0;
  Delay=0;
  Reset=5;
  GPSUpdate=0;
  WriteEEPROMData();

  for(i=0; i<EEMAXSMS; i++) WriteEEPROMSMSNumber(i+1,"");
}

int WriteEEPROMSMSNumber(int index,const char *text){
  return WriteEEPROMField(EENOTIFY+index-1, text);
}

int WriteEEPROMField(int field, const char *text) {
  return WriteEEPROMString( field*EEFIELDSIZE, text, EEFIELDSIZE);
}

int WriteEEPROMIntField(int field, int i) {
char text[EEFIELDSIZE];

  return WriteEEPROMField(field, itoa(i,text,10));
}

int WriteEEPROMString(int offset, const char *text, int maxlen) {
int i;
int c;

  if((offset+maxlen) >= 512) return 0;

  for (i=0; i<maxlen; i++) {
    if(i == (maxlen-1)) EEPROM.write(offset+i, '\0');
    else EEPROM.write(offset+i, text[i]);
    if(text[i] == '\0') break;
  }

#if defined(ESP32)
    EEPROM.commit();
#endif
  
  return i+1;
}

char *ReadEEPROMSMSNumber(int index,char *text){
  ReadEEPROMField(EENOTIFY+index-1, text);
  return text;
}

char *ReadEEPROMField(int field, char *text) {
  strcpy(text,"");
  ReadEEPROMString( field*EEFIELDSIZE, text, EEFIELDSIZE);
  return text;
}

int ReadEEPROMIntField(int field) {
char text[EEFIELDSIZE];

  ReadEEPROMField(field, text);
  return atoi(text);
}

int ReadEEPROMString(int offset, char *text, int maxlen) {
int i;

  if((offset+maxlen) >= 512) return 0;

  for (i=0; i<maxlen; i++) {
    text[i] = EEPROM.read(offset+i);
    if(text[i] == '\0') break;
  }
  
  text[maxlen-1]='\0';

  return i+1;
}
